import logging
import time
import datetime

import ntplib
from dateutil.tz import tzlocal

LOG = logging.getLogger(__name__)

def check_ntp_offset() -> float:
    """
    Sync with NTP server

    :return : offset in seconds
    """
    ntp_servers = [
        'ntp.ntsc.ac.cn', 'ntp.sjtu.edu.cn', 'cn.ntp.org.cn',
        'cn.pool.ntp.org', 'ntp.aliyun.com'
        ]
    retry = len(ntp_servers) - 1
    client = ntplib.NTPClient()
    while retry > 0:
        LOG.info("Try to get time from NTP: %s", ntp_servers[retry])

        try:
            ret = client.request(ntp_servers[retry], version=3)
            offset = (ret.recv_time - ret.orig_time +
                    ret.dest_time - ret.tx_time) / 2
            LOG.info("NTP offset: %.2f seconds", offset)
            return offset
        except ntplib.NTPException:
            LOG.error("Fail to get time, try another")
            retry -= 1
            continue
    return None

def get_timezone():
    """
    Get timezone info
    """
    now_ts_local = time.time()
    now_date_utc   = datetime.datetime.fromtimestamp(now_ts_local, datetime.UTC)
    tl = tzlocal()

    return {
        'tz_name': tl.tzname(now_date_utc),
        'tz_offset': int(tl.utcoffset(now_date_utc).total_seconds()),
    }



import os
import logging
import time
import datetime
import ssl
import json
import requests
import yfinance as yf
import pandas as pd

from .core import FinancialAsset, FinancialMarket

LOG = logging.getLogger(__name__)

STOCK_US_MARKET_ID = "5784f1f5-d8f6-401d-8d24-f685a3812f2d"

class StockUSMarket(FinancialMarket):
    pass

class StockUSAsset(FinancialAsset):

    TYPE_STOCK  = "stock"
    TYPE_ETF    = "etf"
    TYPE_FUTURE = "future"

    def __init__(self, ticker_name:str, market:StockUSMarket,
                 ticker_type=TYPE_STOCK, ticker_cik:int=-1,
                 ticker_title:str=None):
        """
        Constructor

        :param ticker_name  : the name of ticker
        :param market       : the market instance
        :param ticker_type  : the type in stock, etf, future
        :param ticker_cik   : the Central Index Key (CIK)
        """
        super().__init__(ticker_name, market)
        self._ticker_type = ticker_type
        self._ticker_cik = ticker_cik
        self._ticker_title = ticker_title

    @property
    def ticker_type(self):
        return self._ticker_type

    @property
    def asset_type(self):
        return self.ticker_type

    @property
    def ticker_cik(self):
        return self._ticker_cik

    @property
    def ticker_title(self):
        return self._ticker_title

    @property
    def quote(self):
        return "usd"

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret['cik'] = self.ticker_cik
        return ret

class StockUSMarket(FinancialMarket):

    TICKER_LIST_FILE = "stock_us_ticker.json"

    """
    US Stock Market by using Yahoo Financial API
    """

    def __init__(self, cache_dir:str=None):
        """
        :param cache_dir: the root directory for the cache.
        """
        if cache_dir is None:
            cache_dir = os.path.join(os.path.dirname(__file__), "../../cache")
        cache_dir = os.path.join(cache_dir, "StockUS")
        super().__init__("StockUS", "stock", STOCK_US_MARKET_ID, cache_dir)
        self._ready = False
        self._tickers = None

    def milliseconds(self) -> int:
        return round(time.time() * 1000)

    def init(self):
        """
        Initiate the market instance.

        :return: success or not
        """
        if self._ready:
            return False

        ticker_list_path = os.path.join(self.cache_dir, self.TICKER_LIST_FILE)
        if not os.path.exists(ticker_list_path):
            url = "https://www.sec.gov/files/company_tickers.json"
            user_agent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"

            try:
                response = requests.get(url=url,
                                        headers={'User-Agent': user_agent},
                                        timeout=10)
            except requests.exceptions.Timeout:
                LOG.error("request timeout")
                return False

            response.raise_for_status()
            data = response.json()

            # pylint: disable=unspecified-encoding
            with open(ticker_list_path, 'w') as out_file:
                json.dump(data, out_file, sort_keys = True, indent = 4,
                        ensure_ascii = False)

        # pylint: disable=unspecified-encoding
        with open(ticker_list_path, 'r') as input_file:
            ticker_data = json.load(input_file)
            for item in ticker_data.items():
                sa_obj = StockUSAsset(
                    item[1]['ticker'].lower(), self,
                    ticker_cik=item[1]['cik_str'],
                    ticker_title=item[1]['title']
                    )
                self.assets[item[1]['ticker'].lower()] = sa_obj

        LOG.info("Found %d assets for US stock market", len(self.assets))

        self._ready = True
        return True

    def _to_interval(self, timeframe):
        if timeframe in ["1h", "1m", "1d"]:
            return timeframe

        if timeframe == "1M":
            return "1mo"

        if timeframe == "1w":
            return "1wk"

        return None

    def _split_ranges(self, since, to, interval):
        ranges = []
        if to == -1:
            to = time.time()
        index = since
        while index + interval < to:
            ranges.append((index, index + interval))
            index += interval
        ranges.append((index, to))
        LOG.info(ranges)
        return ranges

    def _is_valid_range(self, timeframe, since, to):
        if timeframe == '1m':
            return (to - since) < 8 * 24 * 3600
        if timeframe == '1h':
            return (to - since) < 300 * 24 * 3600
        return True

    def fetch_ohlcv(self, asset:StockUSAsset, timeframe: str, since: int, to: int=-1):
        """
        Fetch OHLCV (Open High Low Close Volume).

        :param     asset: the specific asset
        :param timeframe: 1m/1h/1W/1M etc
        :param     since: the timestamp for starting point
        :param     limit: count
        """
        LOG.info("$$ Fetch from market: timeframe=%s since=%d, to=%d",
                 timeframe, since, to)

        if not self._is_valid_range(timeframe, since, to):
            LOG.error("invalid range")
            return None

        ohlcv = pd.DataFrame()
        try:
            if to != -1:
                ohlcv = yf.download(
                    asset.name,
                    group_by="Ticker",
                    start=datetime.datetime.fromtimestamp(since, tz=datetime.UTC),
                    end=datetime.datetime.fromtimestamp(to, tz=datetime.UTC),
                    interval=self._to_interval(timeframe))
            else:
                ohlcv = yf.download(
                    asset.name,
                    group_by="Ticker",
                    start=datetime.datetime.fromtimestamp(since, tz=datetime.UTC),
                    interval=self._to_interval(timeframe))
            if ohlcv is None or len(ohlcv) == 0:
                return None
        except yf.exceptions.YFPricesMissingError:
            LOG.error("No data for date %s",
                        datetime.datetime.fromtimestamp(since))
            return None
        except ssl.SSLEOFError:
            time.sleep(1)
        except requests.exceptions.SSLError:
            time.sleep(1)

        ohlcv = ohlcv.stack(level=0).rename_axis(
            ['time', 'Ticker']).reset_index(level=1)
        ohlcv = ohlcv[["Open", "High", "Low", "Close", "Volume"]]
        ohlcv.index = pd.to_datetime(ohlcv.index)
        ohlcv.index = ohlcv.index.astype('int64')
        ohlcv.index = ohlcv.index.to_series().div(10**9).astype('int64')
        ohlcv.rename(columns={
            "Open":"open", "High":"high", "Low":"low",
            "Close":"close", "Volume":"vol"}, inplace=True)
        return ohlcv

    def search_ticker(self, company_name:str) -> str:
        """
        Search ticker according to company's name
        """
        url = "https://query2.finance.yahoo.com/v1/finance/search"
        user_agent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"
        params = {"q": company_name, "quotes_count": 1,
                  "country": "United States"}

        try:
            res = requests.get(url=url, params=params,
                            headers={'User-Agent': user_agent},
                            timeout=10)
        except requests.exceptions.Timeout:
            LOG.error("request timeout")
            return None

        data = res.json()

        if 'quotes' not in data or len(data['quotes']) == 0:
            return None

        search_ticker = data['quotes'][0]['symbol']
        asset = self.get_asset(search_ticker)
        if asset is None:
            LOG.error("Could not find the searching ticker %s", search_ticker)
            return None
        return asset



import os
import logging
import time
import json
import ccxt
import pandas as pd
import numpy as np

from .core import FinancialAsset, FinancialMarket
from .timeframe import TimeFrame

LOG = logging.getLogger(__name__)

BINANCE_MARKET_ID = "b13a4902-ad9d-11ef-a239-00155d3ba217"

class CryptoMarket(FinancialMarket):

    def __init__(self, name: str, market_id:str=None, cache_dir:str=None):
        super().__init__(name, FinancialMarket.MARKET_CRYPTO, market_id,
                         cache_dir)

    def get_crypto_asset(self, base, quote):
        """
        Return asset according to crypt naming rule "base/quote"
        """
        return self.get_asset("%s_%s" % (base.lower(), quote.lower()))

class CryptoAsset(FinancialAsset):

    _CRYPTO_TYPE_SPOT = "spot"
    _CRYPTO_TYPE_SWAP = "swap"
    _CRYPTO_TYPE_FUTURE = "future"

    _CRYPTO_TYPES = [ _CRYPTO_TYPE_SPOT, _CRYPTO_TYPE_SWAP, _CRYPTO_TYPE_FUTURE]

    def __init__(self, base:str, quote:str,
                 symbol:str, crypto_type:str, market:CryptoMarket):
        self._base = base.lower()
        self._quote = quote.lower()
        assert crypto_type in self._CRYPTO_TYPES
        self._crypto_type = crypto_type
        self._symbol = symbol
        super().__init__("%s_%s" %(base.lower(),
                                   quote.lower()), market)

    @property
    def base(self) -> str:
        return self._base

    @property
    def quote(self) -> str:
        return self._quote

    @property
    def symbol(self) -> str:
        return self._symbol

    @property
    def crypto_type(self) -> str:
        return self._crypto_type

    @property
    def asset_type(self):
        return self.crypto_type

    def to_dict(self) -> dict:
        retval = super().to_dict()
        retval['base'] = self.base
        retval['symbol'] = self._symbol
        return retval

    def fetch_ohlcv(self, timeframe:str = '1d', since: int = -1, to: int = -1,
                    limit=-1) -> pd.DataFrame:
        """Fetch the OHLCV

        Args:
            timeframe (str, optional): Defaults to '1d'.
            since (int): the start of UTC timestamp.
            to (int, optional): Defaults to -1 for now
            limit (int, optional): the max count of returned value
        """
        df = FinancialAsset.fetch_ohlcv(self, timeframe, since, to, limit)
        LOG.info("Double check whether the dataframe is valid")

        if df is None or len(df) == 0:
            return df

        if not self._cache.check_cache(timeframe, df.index[0], df.index[-1]):
            tfobj = TimeFrame(timeframe)
            (since_new, to_new) = tfobj.normalize(since, to, limit)
            df = self._market.fetch_ohlcv(
                                 self, timeframe, since_new,
                                 to_new)
            self._cache.save(timeframe, df)

        return df

class BinanceMarket(CryptoMarket):

    """
    Binance Market Class to provide crypto information via Binance API.
    Please set the environment variable BINANCE_API_SECRET and BINANCE_API_SECRET.
    """

    _ASSETS_LIST_PATH = "crypto_assets.json"

    def __init__(self, cache_dir:str=None):
        """
        :param cache_dir: the root directory for the cache.
        """
        if cache_dir is None:
            cache_dir = os.path.join(os.path.dirname(__file__), "../../cache")
        cache_dir = os.path.join(cache_dir, "Binance")
        super().__init__("Binance", BINANCE_MARKET_ID, cache_dir)
        assert self.api_key is not None, \
            "Please specify the Binance's API key via the environment" \
            "variable BINANCE_API_KEY"
        assert self.api_secret is not None, \
            "Please specify the Binance's API Secret via the environment" \
            "variable BINANCE_API_SECRET"

        params = {'apiKey': self.api_key, 'secret': self.api_secret}
        if 'HTTP_PROXY' in os.environ:
            params['proxies'] = {
                    'http': os.environ['HTTP_PROXY'],
                    'https': os.environ['HTTP_PROXY']
                    }

        self._ccxt_inst = ccxt.binance(params)
        self._ready = False

    @property
    def api_key(self):
        return os.getenv("BINANCE_API_KEY")

    @property
    def api_secret(self):
        return os.getenv("BINANCE_API_SECRET")

    def milliseconds(self) -> int:
        return self._ccxt_inst.milliseconds()

    def init(self):
        """
        Initiate the market instance.

        :return: success or not
        """
        if self._ready:
            return False

        LOG.info("Loading Crypto Market...")

        asset_list_path = os.path.join(self.cache_dir, self._ASSETS_LIST_PATH)
        if not os.path.exists(asset_list_path):
            LOG.info("Not found the asset list file, use Binance API")
            retry_num = 0
            success = False
            while retry_num < 5:
                retry_num += 1
                try:
                    self._ccxt_inst.load_markets()
                except (ccxt.RequestTimeout, ccxt.NetworkError):
                    LOG.critical("Request Timeout... retry[%d/5]", retry_num)
                    time.sleep(5)
                except Exception as e:
                    LOG.critical(e, exc_info=True)
                    LOG.error("Fail to load market... retry[%d/5]", retry_num)
                    time.sleep(5)
                else:
                    success = True
                    break

            if not success:
                return False

            all_assets = {}
            for symbol in self._ccxt_inst.symbols:
                info = self._ccxt_inst.market(symbol)
                base, quote = symbol.split("/")
                caobj = CryptoAsset(base, quote, symbol, info['type'], self)
                self.assets[caobj.name] = caobj
                all_assets[caobj.name] = caobj.to_dict()

            with open(asset_list_path, 'w', encoding='utf-8') as output:
                json.dump(all_assets, output, sort_keys = True, indent = 4,
                        ensure_ascii = False)
        else:
            LOG.info("Found the asset list file %s", asset_list_path)
            with open(asset_list_path, 'r', encoding='utf-8') as input_file:
                assets_data = json.load(input_file)
                for item in assets_data.items():
                    caobj = CryptoAsset(
                        item[1]['base'],
                        item[1]['quote'],
                        item[1]['symbol'],
                        item[1]['type'],
                        self)
                    self.assets[item[0]] = caobj
        LOG.info("Found %d crypto assets.", len(self.assets))

        self._ready = True
        return True

    def fetch_ohlcv(self, asset:CryptoAsset, timeframe: str, since: int, to:int ):
        """
        Fetch OHLCV (Open High Low Close Volume).

        :param     asset: the specific asset
        :param timeframe: 1m/1h/1W/1M etc
        :param     since: the timestamp for starting point
        :param     limit: count
        """
        LOG.info("$$ Fetch from market: timeframe=%s since=%d, to=%d",
                 timeframe, since, to)

        all_ohlcv = []

        tfobj = TimeFrame(timeframe)
        limit = tfobj.calculate_count(since, to=to)

        remaining = limit
        index = int(since * 1000)
        # Continuous to fetching until get all data
        while remaining > 0:
            LOG.info("since=%d remaining=%d", index, remaining)
            retry = 5
            ohlcv = []
            while retry > 0:
                try:
                    ohlcv = self._ccxt_inst.fetch_ohlcv(asset.symbol, timeframe,
                                                        index, remaining)
                    break
                except TimeoutError:
                    LOG.critical("Network Timeout")
                    time.sleep(1)
                    retry -= 1
                    continue
                except ccxt.NetworkError:
                    LOG.critical("Network Error")
                    time.sleep(1)
                    retry -= 1
                    continue
            all_ohlcv += ohlcv
            if len(ohlcv) <= 1:
                break
            remaining = remaining - len(ohlcv)
            index = ohlcv[-1][0]
            time.sleep(1)

        df = pd.DataFrame(all_ohlcv, columns =
                          ['time', 'open', 'high', 'low', 'close', 'vol'])
        df.time = (df.time / 1000).astype(np.int64)
        df.set_index('time', inplace=True)
        return df

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    cm = BinanceMarket(cache_dir="./binance/")
    cm.init()
    asset_btcusdt = cm.get_crypto_asset("ETH", "USDT")
    while True:
        ret = asset_btcusdt.fetch_ohlcv("1h", limit=3)
        print(ret)
        time.sleep(30)



"""
Time Interval class
"""
import logging
import time
import datetime

LOG = logging.getLogger(__name__)

class TimeFrame:

    SECOND = "s"
    MINUTE = "m"
    HOUR   = "h"
    DAY    = "d"
    WEEK   = "w"
    MONTH  = "M"

    _delta = {
        MINUTE : 60,
        HOUR   : 60 * 60,
        DAY    : 60 * 60 * 24,
        WEEK   : 60 * 60 * 24 * 7
    }

    SUPPORTED = {
        "1m"  : "1min",
        "15m" : "15min",
        "1h"  : "1hour",
        "4h"  : "4hour",
        "8h"  : "8hour",
        "12h" : "12hour",
        "1d"  : "1day",
        "1w"  : "1week",
        "1M"  : "1mon",
    }

    def __init__(self, name="1h") -> None:
        self.interval = name[-1]
        assert self.interval in [ TimeFrame.MINUTE, TimeFrame.HOUR,
            TimeFrame.DAY, TimeFrame.WEEK, TimeFrame.MONTH ]
        self.count = int(name[0:-1])

    def __str__(self) -> str:
        return "%d%s" % (self.count, self.interval)

    def ts_last(self, refer_ts=-1):
        """
        Get the timestamp of the last frame boundary from the reference's
        timestamp. If reference's timestamp is -1, then it is now

        :param current: the end timestamp for reference

        ----------------------------------------------------------
                               ^                            ^
        Time Frame xxxxxxxxxxx |                            |
                     last frame boundary                 current

        """
        if refer_ts == -1:
            refer_ts = time.time()

        if self.interval in [TimeFrame.MINUTE, TimeFrame.HOUR, TimeFrame.DAY]:
            delta_ts = TimeFrame._delta[self.interval] * self.count
            last_now_ts = int(refer_ts / delta_ts) * delta_ts
            return last_now_ts

        today = datetime.datetime.fromtimestamp(refer_ts)
        if self.interval == TimeFrame.WEEK:
            delta = datetime.timedelta(-today.weekday(), weeks=-1 * self.count)
            pre = today + delta
            pre_date = datetime.datetime(pre.year, pre.month, pre.day, 0, 0, 0,
                                         tzinfo=datetime.UTC)
            return pre_date.timestamp()

        if self.interval == TimeFrame.MONTH:
            pre_year = today.year
            pre_month = today.month - 1 * self.count
            if pre_month <= 0:
                pre_month += 12
                pre_year -= 1
            pre_date = datetime.datetime(pre_year, pre_month, 1, 0, 0, 0,
                                         tzinfo=datetime.UTC)
            return pre_date.timestamp()
        return None

    def ts_last_limit(self, limit, to=-1):
        """
        Get the timestamp back in the time before limit count's interval
        till reference timestamp.
        """
        last_ts = self.ts_last(to)
        if self.interval in [TimeFrame.MINUTE, TimeFrame.HOUR,
                             TimeFrame.DAY, TimeFrame.WEEK]:
            delta_ts = TimeFrame._delta[self.interval] * self.count
            return last_ts - (limit - 1) * delta_ts

        if self.interval == TimeFrame.MONTH:
            last_month = datetime.datetime.fromtimestamp(last_ts)
            previous_month_index = last_month.month - (limit - 1)
            previous_year_index = last_month.year

            while previous_month_index < 0:
                previous_month_index += 12
                previous_year_index -= 1
            first_month = datetime.datetime(
                previous_year_index, previous_month_index, 1,
                tzinfo=datetime.UTC)
            return first_month.timestamp()

        return None

    def ts_since(self, since_ts:int) -> int:
        """
        Get the timestamp of the first frame boundary close to since date
        ------------------------------------------------------
           ^             ^
           |             | xxxxxxxxxxx Time Frame  xxxxxxxxxxx
         since    first frame boundary
        if since is just in the first frame boundary, then return it.
        """
        if self.interval in [TimeFrame.MINUTE, TimeFrame.HOUR, TimeFrame.DAY]:
            delta_ts = TimeFrame._delta[self.interval] * self.count
            if int(since_ts) % delta_ts != 0:
                next_ts = (int(since_ts / delta_ts) + 1) * delta_ts
            else:
                next_ts = (int(since_ts / delta_ts)) * delta_ts
            return next_ts

        since_day = datetime.datetime.fromtimestamp(since_ts)
        if since_day.weekday() == 0:
            this_week_first_day_ts = datetime.datetime(
                since_day.year, since_day.month, since_day.day).replace(
                    tzinfo=datetime.timezone.utc).timestamp()
            if this_week_first_day_ts == since_ts:
                return since_ts
        if self.interval == TimeFrame.WEEK:
            next_week_day = since_day + datetime.timedelta(
                days=8 - since_day.isoweekday())
            next_week = datetime.datetime(
                next_week_day.year, next_week_day.month,
                next_week_day.day)
            next_week_ts = next_week.replace(
                tzinfo=datetime.timezone.utc).timestamp()
            return next_week_ts

        if self.interval == TimeFrame.MONTH:
            assert self.count == 1
            this_month_first_day_ts = datetime.datetime(
                since_day.year, since_day.month,
                1).replace(tzinfo=datetime.timezone.utc).timestamp()
            if since_ts == this_month_first_day_ts:
                return since_ts
            if since_day.month == 12:
                next_month = datetime.datetime(
                    since_day.year + 1, 1, 1)
            else:
                next_month = datetime.datetime(
                    since_day.year, since_day.month + 1, 1)

            next_month_ts = next_month.replace(
                tzinfo=datetime.timezone.utc).timestamp()
            return next_month_ts

        return None

    def ts_since_limit(self, since_ts:int, limit:int) -> int:
        next_first_ts = self.ts_since(since_ts)
        if self.interval in [TimeFrame.MINUTE, TimeFrame.HOUR,
                             TimeFrame.DAY, TimeFrame.WEEK]:
            delta_ts = TimeFrame._delta[self.interval] * self.count
            next_last_ts = next_first_ts + (limit - 1) * delta_ts

        if self.interval == TimeFrame.MONTH:
            since_day = datetime.datetime.fromtimestamp(since_ts)
            next_month_index = since_day.month + (limit - 1)
            next_year_index = since_day.year
            while next_month_index >= 12:
                next_month_index -= 12
                next_year_index += 1
            next_month = datetime.datetime(
                next_year_index, next_month_index+1, 1)
            next_last_ts = next_month.replace(
                tzinfo=datetime.timezone.utc).timestamp()

        if next_last_ts > time.time():
            next_last_ts = self.ts_last()

        return next_last_ts

    def calculate_count(self, since:int, max_count:int=-1, to:int=-1) -> int:
        start = self.ts_since(since)
        if to == -1:
            to = self.ts_since_limit(since, max_count)

        assert to >= start, "start:%d to:%d" % (start, to)

        if self.interval in [TimeFrame.MINUTE, TimeFrame.HOUR,
                             TimeFrame.DAY, TimeFrame.WEEK]:
            delta_ts = TimeFrame._delta[self.interval] * self.count
            if max_count != -1:
                return min(max_count, int((to - start) / delta_ts) + 1)
            return int((to - start) / delta_ts) + 1

        if self.interval == TimeFrame.MONTH:
            start_date = datetime.datetime.fromtimestamp(start)
            to_date = datetime.datetime.fromtimestamp(to)
            delta_year = to_date.year - start_date.year
            delta_month = to_date.month - start_date.month
            if delta_month < 0:
                delta_month += 12
                delta_year -= 1
            if max_count == -1:
                return delta_month + delta_year * 12 + 1
            return min(max_count, delta_month + delta_year * 12 + 1)

        return None

    def is_same_frame(self, source, target) -> bool:
        if self.interval in [TimeFrame.MINUTE, TimeFrame.HOUR,
                             TimeFrame.DAY, TimeFrame.WEEK]:
            return abs(target -source) <= \
                TimeFrame._delta[self.interval] * self.count

        if self.interval == TimeFrame.MONTH:
            date_source = datetime.datetime.fromtimestamp(source)
            date_target = datetime.datetime.fromtimestamp(target)
            return (date_source.year == date_target.year and \
                date_target.month == date_source.month)
        return False

    @staticmethod
    def check_valid(tf_str:str):
        return tf_str[-1] in [
            TimeFrame.MINUTE, TimeFrame.HOUR,
            TimeFrame.DAY, TimeFrame.WEEK,
            TimeFrame.MONTH ]

    def normalize(self, since, to, limit):
        if limit == -1:
            assert since != -1, "since must be set without limit"
            since_new = self.ts_since(since)
            to_new = self.ts_last(to)
        else:
            if since == -1:
                since_new = self.ts_last_limit(limit, to)
            else:
                since_new = self.ts_since(since)
            to_new = self.ts_since_limit(since_new, limit)
        return (since_new, to_new)



"""
DataHub Core Package
"""
import os
import logging
import time
import datetime
from abc import ABC, abstractmethod
from threading import Thread
import uuid
import pandas as pd

from .timeframe import TimeFrame

LOG = logging.getLogger(__name__)

class FinancialMarket(ABC):
    # Forward Declaration
    pass

class FinancialAssetCache:
    # Forward Declaration
    pass

class FinancialAsset(ABC):
    """
    Trading instruments are all the different types of assets and contracts that
    can be traded. Trading instruments are classified into various categories,
    some more popular than others.
    """

    def __init__(self, name:str, market:FinancialMarket):
        self._name:str = name
        self._market:FinancialMarket = market
        self._cache:FinancialAssetCache = FinancialAssetCache(self)

    @property
    def name(self) -> str:
        """
        Property name
        """
        return self._name

    @property
    @abstractmethod
    def quote(self) -> str:
        """Quote name, it will be usd for US stock market, cny for chinese stock
        market and usdt for crypto market
        """

    @property
    def market(self) -> FinancialMarket:
        """
        Property market which belong to
        """
        return self._market

    @property
    def cache(self) -> FinancialAssetCache:
        return self._cache

    @property
    @abstractmethod
    def asset_type(self):
        pass

    def to_dict(self) -> dict:
        return {
            'name': self.name,
            'type': self.asset_type,
            'market': self.market.market_id,
            'quote': self.quote
        }

    def fetch_ohlcv(self, timeframe:str = '1d', since: int = -1, to: int = -1,
                    limit=-1) -> pd.DataFrame:
        """Fetch the OHLCV

        Args:
            timeframe (str, optional): Defaults to '1d'.
            since (int): the start of UTC timestamp.
            to (int, optional): Defaults to -1 for now
            limit (int, optional): the max count of returned value
        """
        LOG.info("fetch_ohlcv: since:%d, to:%d, limit:%d tf:%s", since, to,
                 limit, timeframe)
        tfobj = TimeFrame(timeframe)

        # Normalize the since and to, so improve hit ratio in cache
        (since_new, to_new) = tfobj.normalize(since, to, limit)
        LOG.info("Normalize: [%d<->%d] => [%d<->%d]", since, to, since_new,
                 to_new)
        if since_new >= to_new:
            return pd.DataFrame()

        # Case 1: no any data in cache
        cache_start, cache_end = self._cache.get_index(timeframe)
        LOG.info('cache start=%d, end=%d', cache_start, cache_end)
        if cache_start == -1 or cache_end == -1:
            LOG.info("case 1")
            self._cache.save(timeframe, self._market.fetch_ohlcv(
                self, timeframe, since_new, to_new))
        else:
            # Case 2: since < cache_start
            if since_new < cache_start:
                LOG.info("case 2")
                if cache_start > to_new:
                    LOG.warning("The cache will not be continuous after this fetch")
                self._cache.save(timeframe,
                                self._market.fetch_ohlcv(
                                    self, timeframe, since_new,
                                    min(cache_start, to_new)))

            # Case 3: to > cache_end
            if to_new > cache_end:
                LOG.info("case 3")
                if since_new > cache_end:
                    LOG.warning("The cache will not be continuous after this fetch")
                self._cache.save(timeframe,
                                self._market.fetch_ohlcv(
                                    self, timeframe, max(cache_end, since_new),
                                    to_new))

        return self._cache.get_part(timeframe, since_new, to_new)

    def index_to_datetime(self, df:pd.DataFrame, unit="s"):
        df.index = pd.to_datetime(df.index, unit=unit)
        return df

class FinancialMarket(ABC):

    MARKET_CRYPTO = 'crypto'
    MARKET_STOCK = 'stock'

    """
    Trading market includes crypto, stock or golden.
    """

    def __init__(self, name:str, market_type:str, market_id:str=None,
                 cache_dir:str=None):
        assert market_type in \
            [FinancialMarket.MARKET_CRYPTO, FinancialMarket.MARKET_STOCK]
        if market_id is None:
            self._market_id = str(uuid.uuid4())
        else:
            self._market_id = market_id
        self._name = name
        self._assets:dict[str, FinancialAsset] = {}
        self._cache_dir = cache_dir
        self._market_type = market_type

    @property
    def name(self) -> str:
        """
        Property: name
        """
        return self._name

    @property
    def assets(self) -> dict[str, FinancialAsset]:
        """
        Property: assets
        """
        return self._assets

    @property
    def cache_dir(self) -> str:
        """
        Property: Cache Directory
        """
        return self._cache_dir

    @property
    def market_type(self) -> str:
        """
        Property: Market Type
        """
        return self._market_type

    @property
    def market_id(self) -> str:
        return self._market_id


    def get_asset(self, name) -> FinancialAsset:
        """
        Get instrument object from its name
        """
        if name.lower() in self._assets:
            return self._assets[name.lower()]
        return None

    @abstractmethod
    def init(self):
        """
        Financial Market initialization
        """
        raise NotImplementedError

    @abstractmethod
    def fetch_ohlcv(self, asset:FinancialAsset, timeframe:str, since:int, to:int):
        """Fetch OHLCV value for specific asset via Market API

        Args:
            asset (FinancialAsset): _description_
            timeframe (str): _description_
            since (int): _description_
            to (int): _description_

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError

    @abstractmethod
    def milliseconds(self) -> int:
        """
        Current timestamp in milliseconds
        """
        raise NotImplementedError

    def seconds(self) -> int:
        """
        Current timestamp in seconds
        """
        return int(self.milliseconds() / 1000)

class FinancialAssetCache:

    def __init__(self, asset:FinancialAsset):
        self._asset = asset
        self._mem_cache:dict[str, pd.DataFrame] = {}
        self._save_in_progress = False
        self._init()

    def get_index(self, timeframe:str):
        if timeframe not in self._mem_cache:
            self._mem_cache[timeframe] = pd.DataFrame()
            return -1, -1
        cache_obj = self._mem_cache[timeframe]
        if len(cache_obj) == 0:
            return -1, -1
        return cache_obj.index[0], cache_obj.index[-1]

    def get_part(self, timeframe:str, since:int, to:int) -> pd.DataFrame:
        return self._mem_cache[timeframe].loc[since:to]

    def _init(self):
        cache_dir = self._asset.market.cache_dir
        if cache_dir is None or not os.path.exists(cache_dir):
            return

        for name, _ in TimeFrame.SUPPORTED.items():
            csv_name = self._get_csv_name(name)
            csv_path = os.path.join(cache_dir, csv_name)
            if os.path.exists(csv_path):
                LOG.info("found: %s", csv_path)
                try:
                    self._mem_cache[name] = \
                        pd.read_csv(csv_path, index_col=0)
                except pd.errors.EmptyDataError:
                    LOG.info("Found blank file %s", csv_path)

    def search(self, timeframe:str, since:int, to:int):
        """
        Search from cache
        """
        LOG.info("Search cache: tf=%s, since=%d, to=%d",
                 timeframe, since, to)
        if timeframe not in self._mem_cache:
            self._mem_cache[timeframe] = pd.DataFrame()
            return None

        if len(self._mem_cache[timeframe]) == 0:
            return None

        if since < self._mem_cache[timeframe].index[0] or \
            since > self._mem_cache[timeframe].index[-1]:
            LOG.info("No records found from cache")
            return None

        df_part = None
        # if from_ in the range of existing cache
        if to <= self._mem_cache[timeframe].index[-1] and \
            self.check_cache(timeframe, since, to):
            LOG.info("All records found from cache")
            df_part = self._mem_cache[timeframe].loc[since:to]
        else:
            if self.check_cache(timeframe, since):
                df_part = self._mem_cache[timeframe].loc[since:]
                LOG.info("Part of records found from cache: from %d -> %d",
                        df_part.index[0], df_part.index[-1])

        return df_part

    def save(self, timeframe:str, df_new:pd.DataFrame):
        """
        Save OHLCV to cache
        """
        if df_new is None or len(df_new) == 0:
            LOG.warning("Invalid dataframe for save")
            return
        self._mem_cache[timeframe] = pd.concat(
            [self._mem_cache[timeframe], df_new])
        self._mem_cache[timeframe] = \
            self._mem_cache[timeframe][~self._mem_cache[timeframe].\
                                       index.duplicated(keep='first')]
        self._mem_cache[timeframe].sort_index(inplace=True)
        self._save_cache_to_file(timeframe)

    def _get_csv_name(self, timeframe):
        return self._asset.name + "-" + TimeFrame.SUPPORTED[timeframe] + ".csv"

    def _save_cache_to_file(self, timeframe):
        self._save_in_progress = True
        cache_dir = self._asset.market.cache_dir
        if cache_dir is not None:
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir)
            fname = os.path.join(self._asset.market.cache_dir,
                                self._get_csv_name(timeframe))
            self._mem_cache[timeframe].to_csv(fname)
            LOG.info("save to file: %s", fname)
        self._save_in_progress = False

    def check_cache(self, timeframe:str, since:int, to:int=-1):
        """
        Check whether the dataframe is continuous
        """
        if timeframe not in self._mem_cache:
            self._mem_cache[timeframe] = pd.DataFrame()
            return False

        df_cached = self._mem_cache[timeframe]
        if len(df_cached) == 0:
            return False

        if to == -1:
            to = self._mem_cache[timeframe].index[-1]

        for item in [since, to]:
            if item < df_cached.index[0] or item > df_cached.index[-1]:
                return False

        df_cached = self._mem_cache[timeframe].loc[since:to]
        if len(df_cached) == 0 or df_cached.index[0] != since:
            return False

        tfobj = TimeFrame(timeframe)
        count = tfobj.calculate_count(
            since=df_cached.index[0], to=df_cached.index[-1])
        if count != len(df_cached):
            LOG.error("The cache[%d->%d] is not completed: count=%d, len=%d",
                       since, to, count, len(df_cached))
            return False
        return True

class DataCollectorThread(Thread):

    def __init__(self, key:str, market_obj:FinancialMarket,
                 asset_obj:FinancialAsset, timeframe:str, since:int):
        Thread.__init__(self)
        self._key = key
        self._market_obj = market_obj
        self._since = since
        self._asset_obj = asset_obj
        self._timeframe = timeframe
        self._current = since
        self._terminate = False
        self._now = time.time()

    def run(self):
        LOG.info("Thread %s started.", self._key)
        self._current = self._since
        limit = -1
        tfobj = TimeFrame(self._timeframe)

        while not self._terminate:
            LOG.info("=> %d: Collector[%s] since=%d ...",
                 self._now, datetime.datetime.fromtimestamp(self._now).\
                    strftime('%Y-%m-%d %H:%M:%S'),
                    self._current)
            if tfobj.is_same_frame(self._current, self._now):
                break

            to = tfobj.ts_since_limit(self._current, limit)
            if self._asset_obj.cache.check_cache(
                self._timeframe, self._current, to):
                # skip for existing data
                LOG.info("Skip the range [%d->%d] since already in cache.",
                         self._current, to)
                self._current = tfobj.ts_since_limit(to + 1, limit)
                continue

            ret = self._asset_obj.fetch_ohlcv(
                self._timeframe, self._current, limit=limit)
            if ret is not None:
                if len(ret) <= 1:
                    break
                self._current = tfobj.ts_since_limit(ret.index[-1] + 1, 1)
                LOG.info("current:%d, now:%d", self._current, self._now)
                if tfobj.is_same_frame(self._current, self._now):
                    break
            else:
                break

            time.sleep(5)
        self._terminate = True
        LOG.info("Thread %s completed.", self._key)

    @property
    def is_completed(self):
        return self._terminate

    def terminate(self):
        self._terminate = True

    @property
    def progress(self):
        tfobj = TimeFrame(self._timeframe)
        total = tfobj.calculate_count(since=self._since, to=self._now)
        now = tfobj.calculate_count(since=self._current, to=self._now)
        return now, total

    @property
    def since(self):
        return self._since



"""
The collections of basic strategies
"""
import logging

import backtrader as bt
import backtrader.indicators as btind

LOG = logging.getLogger(__name__)

# pylint: disable=no-member, too-many-function-args, unexpected-keyword-arg

class StrategyBase(bt.Strategy):

    def log(self, txt, dt=None):
        dt = dt or self.datas[0].datetime.date(0)
        LOG.info('%s %s', dt.strftime("%Y-%m-%d %H:%M:%S"), txt)

    def notify_order(self, order):
        if order.status in [order.Submitted, order.Accepted]:
        # Buy/Sell order submitted/accepted to/by broker - Nothing to do
            return

        # Check if an order has been completed
        # Attention: broker could reject order if not enough cash
        if order.status in [order.Completed]:
            if order.isbuy():
                self.log('BUY EXECUTED, %.2f' % order.executed.price)
            elif order.issell():
                self.log('SELL EXECUTED, %.2f' % order.executed.price)
            self.bar_executed = len(self)
        elif order.status in [order.Canceled, order.Margin, order.Rejected]:
            self.log('Order Canceled/Margin/Rejected')

        # Write down: no pending order
        self.order = None

    def notify_trade(self, trade):
        if not trade.isclosed:
            return

        self.log('OPERATION PROFIT, GROSS %.2f, NET %.2f' % (
            trade.pnl, trade.pnlcomm))

class StrategyRsi(StrategyBase):
    params=(
        ('min_RSI', 35),
        ('max_RSI', 65),
        ('max_position', 10),
        ('look_back_period', 14)
        )

    def __init__(self):
        super().__init__()
        # RSI indicator
        self.RSI = btind.RSI_SMA(self.data.close, period=self.params.look_back_period)

    def next(self):
        # Buy if over sold
        if self.RSI < self.params.min_RSI:
            self.buy()
        # Sell if over buyed
        if self.RSI > self.params.max_RSI:
            self.close()

class StrategySma(StrategyBase):

    """
    SMA, SimpleMovingAverage
    Non-weighted average of the last n periods

    Formula:
        movav = Sum(data, period) / period

    See also:
        http://en.wikipedia.org/wiki/Moving_average#Simple_moving_average
    """

    params = (
        ('fast', 9),
        ('slow', 21),
    )

    def __init__(self):
        super().__init__()
        sma_fast = btind.SMA(period=self.p.fast)
        sma_slow = btind.SMA(period=self.p.slow)
        self.buysig = btind.CrossOver(sma_fast, sma_slow)

    def next(self):
        if self.position.size:
            if self.buysig < 0:
                self.sell()
        elif self.buysig > 0:
            self.buy()

class StrategyMacd(StrategyBase):
    params=(
        ('fast_LBP',12),
        ('slow_LBP',26),
        ('max_position',1),
        ('signal_LBP',9)
        )

    def log(self, txt, dt=None):
        ''' Logging function for this strategy'''
        dt = dt or self.datas[0].datetime.date(0)
        print('%s, %s' % (dt.isoformat(), txt))

    def __init__(self):
        super().__init__()
        self.fast_EMA = btind.EMA(self.data, period=self.params.fast_LBP)
        self.slow_EMA = btind.EMA(self.data, period=self.params.slow_LBP)

        self.MACD=self.fast_EMA-self.slow_EMA
        self.Signal = btind.EMA(self.MACD, period=self.params.signal_LBP)
        self.Crossing = btind.CrossOver(
            self.MACD,
            self.Signal,
            plotname='Buy_Sell_Line')
        self.Hist = self.MACD - self.Signal

    def next(self):

        # If MACD is above Signal line
        if self.Crossing > 0:
            if self.position.size < self.params.max_position:
                self.buy()

        # If MACD is below Signal line
        elif self.Crossing < 0:
            if self.position.size > 0:
                self.close()

class StrategyWma(StrategyBase):

    """
    Alias:

        WMA, MovingAverageWeighted
        A Moving Average which gives an arithmetic weighting to values with the
        newest having the more weight

    Formula:

        weights = range(1, period + 1)
        coef = 2 / (period * (period + 1))
        movav = coef * Sum(weight[i] * data[period - i] for i in range(period))

    See also:

        http://en.wikipedia.org/wiki/Moving_average#Weighted_moving_average
    """

    params = (
        ('maperiod', 30),
    )

    def __init__(self):
        super().__init__()
        self.dataclose = self.datas[0].close

        # To keep track of pending orders and buy price/commission
        self.order = None
        self.buyprice = None
        self.buycomm = None

        #Adding SMA indicator
        self.sma = btind.WeightedMovingAverage(
            self.datas[0], period = self.params.maperiod
        )

    def next(self):
        self.log('Close, %.2f' % self.dataclose[0])

        if self.order:
            return

        #check if we are in market
        if not self.position:
            if self.dataclose[0] > self.sma[0]:
                self.log('BUY CREATE, %.2f' % self.dataclose[0])
                self.order = self.buy()
        else:
            if self.dataclose[0] < self.sma[0]:
                self.log('SELL CREATE, %.2f' % self.dataclose[0])
                self.order = self.sell()


class StrategyBb(StrategyBase):

    """
    BollingerBands

    Defined by John Bollinger in the 80s. It measures volatility by defining
    upper and lower bands at distance x standard deviations

    Formula:
        midband = SimpleMovingAverage(close, period)
        topband = midband + devfactor * StandardDeviation(data, period)
        botband = midband - devfactor * StandardDeviation(data, period)

    See: <http://en.wikipedia.org/wiki/Bollinger_Bands>
    """
    params = (
        ('maperiod', 30 ),
    )

    def __init__(self):
        super().__init__()
        self.dataclose = self.datas[0].close

        # To keep track of pending orders and buy price/commission
        self.order = None
        self.buyprice = None
        self.buycomm = None

        #Adding SMA indicator
        self.bbands = btind.BBands(self.datas[0])

    def next(self):
        self.log('Close, %.2f' % self.dataclose[0])

        if self.order:
            return

        #check if we are in market
        if not self.position:
            if self.bbands[0] < self.dataclose[0]:
                self.log('BUY CREATE, %.2f' % self.dataclose[0])
                self.order = self.buy()
        else:
            if self.bbands[0] > self.dataclose[0]:
                self.log('SELL CREATE, %.2f' % self.dataclose[0])
                self.order = self.sell()
